from turbot.blocks import (
    section_block, text_block, button_block, accessory_block
)
from turbot.channel import channel_url
from boto3.dynamodb.conditions import Key
import turbot.sheets
import re

def find_puzzle_for_sort_key(turb, hunt_id, sort_key):
    """Given a hunt_id and puzzle_id, return that puzzle

    Returns None if no puzzle with the given hunt_id and puzzle_id
    exists in the database, otherwise a dictionary with all fields
    from the puzzle's row in the database.
    """

    response = turb.table.get_item(
        Key={
            'hunt_id': hunt_id,
            'SK': sort_key,
        })

    if 'Item' in response:
        return response['Item']
    else:
        return None

def find_puzzle_for_url(turb, hunt_id, url):
    """Given a hunt_id and URL, return the puzzle with that URL

    Returns None if no puzzle with the given URL exists in the database,
    otherwise a dictionary with all fields from the puzzle's row in
    the database.
    """

    response = turb.table.query(
        IndexName='url_index',
        KeyConditionExpression=(
            Key('hunt_id').eq(hunt_id) &
            Key('url').eq(url)
        )
    )

    if response['Count'] == 0:
        return None

    return response['Items'][0]

def puzzle_blocks(puzzle, include_rounds=False):
    """Generate Slack blocks for a puzzle

    The puzzle argument should be a dictionary as returned from the
    database. The return value can be used in a Slack command
    expecting blocks to provide all the details of a puzzle, (its
    state, solution, links to channel and sheet, etc.).
    """

    name = puzzle['name']
    status = puzzle['status']
    solution = puzzle['solution']
    channel_id = puzzle['channel_id']
    url = puzzle.get('url', None)
    sheet_url = puzzle.get('sheet_url', None)
    state = puzzle.get('state', None)
    tags = puzzle.get('tags', [])
    status_emoji = ''
    solution_str = ''

    if status == 'solved':
        status_emoji = ":ballot_box_with_check:"
    else:
        status_emoji = ":white_square:"

    if len(solution):
        solution_str = "*`" + '`, `'.join(solution) + "`*"

    meta_str = ''
    if puzzle.get('type', 'plain') == 'meta':
        meta_str = "*META* "

    links = []
    if url:
        links.append("<{}|Puzzle>".format(url))
    if sheet_url:
        links.append("<{}|Sheet>".format(sheet_url))

    state_str = ''
    if state:
        state_str = " State: {}".format(state)

    tags_str = ''
    if tags:
        tags_str = " Tags: "+" ".join(["`{}`".format(tag) for tag in tags])

    extra_str = ''
    if state_str or tags_str:
        extra_str = "\n{}{}".format(tags_str, state_str)

    rounds_str = ''
    if include_rounds and 'rounds' in puzzle:
        rounds = puzzle['rounds']
        rounds_str = " in round{}: {}".format(
            "s" if len(rounds) > 1 else "",
            ", ".join(rounds)
        )

    puzzle_text = "{} {}<{}|{}> {} ({}){}{}".format(
        status_emoji,
        meta_str,
        channel_url(channel_id), name,
        solution_str,
        ', '.join(links), rounds_str,
        extra_str
    )

    # Combining hunt ID and puzzle ID together here is safe because
    # hunt_id is restricted to not contain a hyphen, (see
    # valid_id_re in interaction.py)
    hunt_and_sort_key = "{}-{}".format(puzzle['hunt_id'], puzzle['SK'])

    return [
        accessory_block(
            section_block(text_block(puzzle_text)),
            button_block("✏", "edit_puzzle", hunt_and_sort_key)
        )
    ]

def puzzle_matches_one(puzzle, pattern):
    """Returns True if this puzzle matches the given string (regexp)

    A match will be considered on any of puzzle title, round title,
    puzzle URL, puzzle state, puzzle type, tags, or solution
    string. The string can include regular expression syntax. Matching
    is case insensitive.
    """

    p = re.compile('.*'+pattern+'.*', re.IGNORECASE)

    if p.match(puzzle['name']):
        return True

    if 'rounds' in puzzle:
        for round in puzzle['rounds']:
            if p.match(round):
                return True

    if 'url' in puzzle:
        if p.match(puzzle['url']):
            return True

    if 'state' in puzzle:
        if p.match(puzzle['state']):
            return True

    if 'type' in puzzle:
        if p.match(puzzle['type']):
            return True

    if 'solution' in puzzle:
        for solution in puzzle['solution']:
            if p.match(solution):
                return True

    if 'tags' in puzzle:
        for tag in puzzle['tags']:
            if p.match(tag):
                return True

    return False

def puzzle_matches_all(puzzle, patterns):
    """Returns True if this puzzle matches all of the given list of patterns

    A match will be considered on any of puzzle title, round title,
    puzzle URL, puzzle state, puzzle types, tags, or solution
    string. All patterns must match the puzzle somewhere, (that is,
    there is an implicit logical AND between patterns). Patterns can
    include regular expression syntax. Matching is case insensitive.

    """

    for pattern in patterns:
        if not puzzle_matches_one(puzzle, pattern):
            return False

    return True

def puzzle_id_from_name(name):
    return re.sub(r'[^a-zA-Z0-9_]', '', name).lower()

def puzzle_sort_key(puzzle):
    """Return an appropriate sort key for a puzzle in the database

    The sort key must start with "puzzle-" to distinguish puzzle items
    in the database from all non-puzzle items. After that, though, the
    only requirements are that each puzzle have a unique key and they
    give us the ordering we want. And for ordering, we want meta puzzles
    before non-meta puzzles and then alphabetical order by name within
    each of those groups.

    So puting a "-meta-" prefix in front of the puzzle ID does the trick.
    """

    return "puzzle-{}{}".format(
        "-meta-" if puzzle['type'] == "meta" else "",
        puzzle['puzzle_id']
    )

def puzzle_channel_topic(puzzle):
    """Compute the channel topic for a puzzle"""

    topic = ''

    if puzzle['status'] == 'solved':
        topic += "SOLVED: `{}` ".format('`, `'.join(puzzle['solution']))

    topic += puzzle['name']

    links = []

    url = puzzle.get('url', None)
    if url:
        links.append("<{}|Puzzle>".format(url))

    sheet_url = puzzle.get('sheet_url', None)
    if sheet_url:
        links.append("<{}|Sheet>".format(sheet_url))

    if len(links):
        topic += "({})".format(', '.join(links))

    tags = puzzle.get('tags', [])
    if tags:
        topic += " {}".format(" ".join(["`{}`".format(t) for t in tags]))

    state = puzzle.get('state', None)
    if state:
        topic += " {}".format(state)

    return topic

def puzzle_channel_name(puzzle):
    """Compute the channel name for a puzzle"""

    round = ''
    if 'rounds' in puzzle:
        round = '-' + puzzle_id_from_name(puzzle['rounds'][0])

    meta = ''
    if puzzle['type'] == 'meta':
        meta = '-m'

    # Note: We don't use puzzle['puzzle_id'] here because we're keeping
    # that as a persistent identifier in the database. Instead we
    # create a new ID-like identifier from the current name.
    channel_name = "{}{}{}-{}".format(
        puzzle['hunt_id'],
        round,
        meta,
        puzzle_id_from_name(puzzle['name'])
    )

    if puzzle['status'] == 'solved':
        channel_name += "-solved"

    return channel_name

def puzzle_sheet_name(puzzle):
    """Compute the sheet name for a puzzle"""

    sheet_name = puzzle['name']
    if puzzle['status'] == 'solved':
        sheet_name += " - Solved {}".format(", ".join(puzzle['solution']))

    return sheet_name

def puzzle_update_channel_and_sheet(turb, puzzle, old_puzzle=None):

    channel_id = puzzle['channel_id']

    # Compute the channel topic and set it if it has changed
    channel_topic = puzzle_channel_topic(puzzle)

    old_channel_topic = None
    if old_puzzle:
        old_channel_topic = puzzle_channel_topic(old_puzzle)

    if channel_topic != old_channel_topic:
        # Slack only allows 250 characters for a topic
        if len(channel_topic) > 250:
            channel_topic = channel_topic[:247] + "..."
        turb.slack_client.conversations_setTopic(channel=channel_id,
                                                 topic=channel_topic)

    # Compute the sheet name and set it if it has changed
    sheet_name = puzzle_sheet_name(puzzle)

    old_sheet_name = None
    if old_puzzle:
        old_sheet_name = puzzle_sheet_name(old_puzzle)

    if sheet_name != old_sheet_name:
        turbot.sheets.rename_spreadsheet(turb, puzzle['sheet_url'], sheet_name)

    # Compute the Slack channel name and set it if it has changed
    channel_name = puzzle_channel_name(puzzle)

    old_channel_name = None
    if old_puzzle:
        old_channel_name = puzzle_channel_name(old_puzzle)

    if channel_name != old_channel_name:
        turb.slack_client.conversations_rename(
            channel=channel_id,
            name=channel_name
        )

# A copy deep enough to work for puzzle_update_channel_and_sheet above
def puzzle_copy(old_puzzle):
    new_puzzle = old_puzzle.copy()

    if 'tags' in old_puzzle:
        new_puzzle['tags'] = old_puzzle['tags'].copy()

    return new_puzzle
